@@ -3,6 +3,7 @@
 
 #include <vector>
 #include <map>
+#include <algorithm>
 #include <string.h>
 #include <memory.h>
 
@@ -19,7 +20,7 @@
 #include <qpdf/QPDF_Null.hh>
 #include <qpdf/QPDF_Dictionary.hh>
 
-std::string QPDF::qpdf_version = "5.0.0";
+std::string QPDF::qpdf_version = "5.1.0";
 
 static char const* EMPTY_PDF =
     "%PDF-1.3\n"
@@ -486,7 +487,7 @@ QPDF::read_xref(qpdf_offset_t xref_offset)
 qpdf_offset_t
 QPDF::read_xrefTable(qpdf_offset_t xref_offset)
 {
-    PCRE xref_first_re("^\\s*(\\d+)\\s+(\\d+)");
+    PCRE xref_first_re("^\\s*(\\d+)\\s+(\\d+)\\s*");
     PCRE xref_entry_re("(?s:(^\\d{10}) (\\d{5}) ([fn])[ \r\n]{2}$)");
 
     std::vector<QPDFObjGen> deleted_items;
@@ -495,7 +496,10 @@ QPDF::read_xrefTable(qpdf_offset_t xref_offset)
     bool done = false;
     while (! done)
     {
-	std::string line = this->file->readLine(50);
+        char linebuf[51];
+        memset(linebuf, 0, sizeof(linebuf));
+        this->file->read(linebuf, sizeof(linebuf) - 1);
+	std::string line = linebuf;
 	PCRE::Match m1 = xref_first_re.match(line.c_str());
 	if (! m1)
 	{
@@ -504,6 +508,8 @@ QPDF::read_xrefTable(qpdf_offset_t xref_offset)
 			  "xref table", this->file->getLastOffset(),
 			  "xref syntax invalid");
 	}
+        file->seek(this->file->getLastOffset() + m1.getMatch(0).length(),
+                   SEEK_SET);
 	int obj = atoi(m1.getMatch(1).c_str());
 	int num = atoi(m1.getMatch(2).c_str());
 	static int const xref_entry_size = 20;
@@ -531,7 +537,7 @@ QPDF::read_xrefTable(qpdf_offset_t xref_offset)
             // For xref_table, these will always be small enough to be ints 
 	    qpdf_offset_t f1 = QUtil::string_to_ll(m2.getMatch(1).c_str());
 	    int f2 = atoi(m2.getMatch(2).c_str());
-	    char type = m2.getMatch(3)[0];
+	    char type = m2.getMatch(3).at(0);
 	    if (type == 'f')
 	    {
 		// Save deleted items until after we've checked the
@@ -699,7 +705,26 @@ QPDF::processXRefStream(qpdf_offset_t xref_offset, QPDFObjectHandle& xref_obj)
 		      "Cross-reference stream does not have"
 		      " proper /W and /Index keys");
     }
-    std::vector<int> indx;
+
+    int W[3];
+    size_t entry_size = 0;
+    int max_bytes = sizeof(qpdf_offset_t);
+    for (int i = 0; i < 3; ++i)
+    {
+	W[i] = W_obj.getArrayItem(i).getIntValue();
+        if (W[i] > max_bytes)
+        {
+            throw QPDFExc(qpdf_e_damaged_pdf, this->file->getName(),
+                          "xref stream", xref_offset,
+                          "Cross-reference stream's /W contains"
+                          " impossibly large values");
+        }
+	entry_size += W[i];
+    }
+    long long max_num_entries =
+        static_cast<unsigned long long>(-1) / entry_size;
+
+    std::vector<long long> indx;
     if (Index_obj.isArray())
     {
 	int n_index = Index_obj.getArrayNItems();
@@ -731,25 +756,29 @@ QPDF::processXRefStream(qpdf_offset_t xref_offset, QPDFObjectHandle& xref_obj)
     else
     {
 	QTC::TC("qpdf", "QPDF xref /Index is null");
-	int size = dict.getKey("/Size").getIntValue();
+	long long size = dict.getKey("/Size").getIntValue();
 	indx.push_back(0);
 	indx.push_back(size);
     }
 
-    int num_entries = 0;
+    long long num_entries = 0;
     for (unsigned int i = 1; i < indx.size(); i += 2)
     {
-	num_entries += indx[i];
-    }
-
-    int W[3];
-    int entry_size = 0;
-    for (int i = 0; i < 3; ++i)
-    {
-	W[i] = W_obj.getArrayItem(i).getIntValue();
-	entry_size += W[i];
+        if (indx.at(i) > max_num_entries - num_entries)
+        {
+            throw QPDFExc(qpdf_e_damaged_pdf, this->file->getName(),
+                          "xref stream", xref_offset,
+                          "Cross-reference stream claims to contain"
+                          " too many entries: " +
+                          QUtil::int_to_string(indx.at(i)) + " " +
+                          QUtil::int_to_string(max_num_entries) + " " +
+                          QUtil::int_to_string(num_entries));
+        }
+	num_entries += indx.at(i);
     }
 
+    // entry_size and num_entries have both been validated to ensure
+    // that this multiplication does not cause an overflow.
     size_t expected_size = entry_size * num_entries;
 
     PointerHolder<Buffer> bp = xref_obj.getStreamData();
@@ -777,6 +806,9 @@ QPDF::processXRefStream(qpdf_offset_t xref_offset, QPDFObjectHandle& xref_obj)
 
     bool saw_first_compressed_object = false;
 
+    // Actual size vs. expected size check above ensures that we will
+    // not overflow any buffers here.  We know that entry_size *
+    // num_entries is equal to the size of the buffer.
     unsigned char const* data = bp->getBuffer();
     for (int i = 0; i < num_entries; ++i)
     {
@@ -803,9 +835,9 @@ QPDF::processXRefStream(qpdf_offset_t xref_offset, QPDFObjectHandle& xref_obj)
 	// based on /Index.  The generation number is 0 unless this is
 	// an uncompressed object record, in which case the generation
 	// number appears as the third field.
-	int obj = indx[cur_chunk] + chunk_count;
+	int obj = indx.at(cur_chunk) + chunk_count;
 	++chunk_count;
-	if (chunk_count >= indx[cur_chunk + 1])
+	if (chunk_count >= indx.at(cur_chunk + 1))
 	{
 	    cur_chunk += 2;
 	    chunk_count = 0;
